import logging


logger = logging.getLogger("hyperbolic")
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
ch.setFormatter(
    logging.Formatter('%(levelname)s:%(name)s:%(message)s'))
logger.addHandler(ch)



#!/usr/bin/env python3
import argparse

import numpy as np
import pandas as pd

import hyperbolic
import hyperbolic.config
import hyperbolic.plots


parser = argparse.ArgumentParser(
    description="Compute the hyperbolic magnitudes and add them into the "
                "input data catalogue.",
    add_help=False)
parser.add_argument(
    "infile", metavar="infile",
    help="input FITS file")
parser.add_argument(
    "--hdu", type=int, default=1,
    help="FITS HDU index to read (default: %(default)s)")
parser.add_argument(
    "outfile", metavar="outfile",
    help="output path for FITS file with added hyperbolic magnitudes")

parser.add_argument(
    "-c", "--config", required=True,
    help="JSON configuration file that specifies in- and output data "
         "(see --dump)")
parser.add_argument(
    "-s", "--stats", required=True,
    help="statistics file genereated with 'hyp_smoothing.py'")
parser.add_argument(
    "--smoothing",
    help="external statistics file genereated with 'hyp_smoothing.py' that is "
         "used to define the smoothing parameter b")
parser.add_argument(
    "-f", "--fields",
    help="column name that can uniquely indentify pointings")
parser.add_argument(
    "--b-global", action='store_true',
    help="compute the smoothing paramter globally for all filters")
parser.add_argument(
    "--plot", action='store_true',
    help="add a PDF file with summary plots alongside the output file")

group = parser.add_argument_group("help")
group.add_argument(
    "-h", "--help", action="help",
    help="show this help message and exit")
group.add_argument(
    "-d", "--dump", nargs=0, action=hyperbolic.config.DumpAction,
    help="dump an empty configuration file, 'filter_name' can be repeated")
parser.add_argument(
    "-v", "--verbose", action="store_true",
    help="show statistic summary per filter")


if __name__ == "__main__":
    config = hyperbolic.config.LoadConfigMagnitudes(parser.parse_args())
    # load the input data
    data = config.load_input()
    all_stats = config.load_stats()  # for the internal zeropoint corrections
    smooth_stats = config.load_smoothing()
    # get data columns
    fields = config.get_fields(data)
    fluxes = config.get_fluxes(data)
    errors = config.get_errors(data)

    # compute smoothing factor (from external data if --smoothing is provided)
    b = smooth_stats[hyperbolic.Keys.b].groupby(
        hyperbolic.Keys.filter).agg(np.nanmedian)
    if config.b_global:
        b[:] = np.median(b)
    if config.verbose:  # print in nicely looking dataframe
        _df = pd.DataFrame({hyperbolic.Keys.b: b})
        print(_df.loc[config.filters])  # maintain usual order
    b = b.to_dict()

    for filt in config.filters:
        hyperbolic.logger.logger.info(f"processing filter {filt}")
        stats = hyperbolic.fill_missing_stats(all_stats.loc[filt])
        # mask to valid observations
        is_good = errors[filt] > 0.0

        # compute normalised flux
        ref_flux = hyperbolic.fields_to_source(
            stats[hyperbolic.Keys.ref_flux], fields, index=fluxes[filt].index)
        norm_flux = fluxes[filt] / ref_flux
        norm_flux_err = errors[filt] / ref_flux

        # compute the hyperbolic magnitudes
        hyp_mag = hyperbolic.compute_magnitude(
            norm_flux, b[filt])
        hyp_mag_err = hyperbolic.compute_magnitude_error(
            norm_flux, b[filt], norm_flux_err)

        # add data to catalogue
        key_mag = config.outname[filt]
        key_mag_err = config.KiDS_aware_error_colname(key_mag)
        data[key_mag] = np.where(is_good, hyp_mag, -99.0)
        data[key_mag_err] = np.where(is_good, hyp_mag_err, -99.0)

    if config.plot:
        with hyperbolic.plots.Plotter(config, data, all_stats, b) as plotter:
            if config.fields is not None:
                plotter.plot_b(smooth_stats)
            plotter.plot_magnitudes()
            plotter.plot_magnitude_distribution()
            plotter.plot_colour_distribution()
            plotter.plot_magdiff()
    config.write_output(data)



import warnings
from typing import Optional

import numpy as np
import numpy.typing as npt
import pandas as pd

from . import Keys
from .logger import logger


pogson = 2.5 * np.log10(np.e)


def fields_to_source(
    per_field_data: pd.Series,
    fields: npt.NDArray,
    index: Optional[pd.Index] = None
) -> pd.Series:
    """
    Map field-wide statistics back to each individual source based on the field membership.

    Parameters:
    -----------
    per_field_data : pd.Series
        Series with field identifier as index and statistics as values.
    fields : array-like
        List of field identifiers that defines the field membership of each source.
    index : pd.Index
        Optional index to assign to the output.
    
    Returns:
    --------
    per_source_data : pd.Series
        Field-wide statistics mapped to each source.
    """
    per_source_data = per_field_data.loc[fields]
    if index is not None:
        per_source_data.index = index
    return per_source_data


def ref_flux_from_zp(m_0: npt.ArrayLike) -> npt.ArrayLike:
    """
    Convert the photometric zeropoint to the reference flux.
    """
    return np.exp(m_0 / pogson)


def zp_from_ref_flux(flux: npt.ArrayLike) -> npt.ArrayLike:
    """
    Convert the reference flux to the photometric zeropoint
    """
    return pogson * np.log(flux)


def estimate_zp(
    magnitude: npt.ArrayLike,
    flux: npt.ArrayLike
) -> npt.ArrayLike:
    """
    Estimate the photometric zeropoint from a set of magnitudes and corresponding fluxes.

    Parameters:
    -----------
    magnitude : array-like
        Magnitudes of objects.
    flux : array-like
        Fluxes of objects.
    
    Returns:
    --------
    zp : array-like
        Photometric zeropoint per object.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return magnitude + pogson * np.log(flux)


def convert_flux(
    flux: npt.ArrayLike,
    zeropoint: npt.ArrayLike,
    target_zp: npt.ArrayLike
) -> npt.ArrayLike:
    """
    Convert a flux measurement obtained with a given photometric zeropoint to a flux system with a
    different photometric zeropoint.

    Parameters:
    -----------
    flux : array-like
        Input fluxes of objects.
    zeropoint : array-like
        Photometric zeropoint (per source) corresponding to the input fluxes.
    target_zp : array-like
        Photometric zeropoint (per source) for the output fluxes.
    
    Returns:
    --------
    flux : array-like
        Output fluxes for the new photometric zeropoint.
    """
    return flux * np.exp((target_zp - zeropoint) / pogson)


def estimate_b(
    zp: npt.ArrayLike,
    flux_error: npt.ArrayLike
) -> npt.ArrayLike:
    """
    Estimate the smoothing factor for hyperbolic magnitudes in flux units for a given photometric
    zeropoint and flux errors.

    Parameters:
    -----------
    zp : array-like
        Photometric zeropoint (per source) for the output fluxes.
    flux_error : array-like
        Flux error (per object).

    Returns:
    --------
    b : array-like
        Estimate of the smoothing factor (per object).
    """
    return np.sqrt(pogson) * np.exp(-zp / pogson) * flux_error


def compute_classic_magnitude(
    flux: npt.ArrayLike,
    zeropoint: npt.ArrayLike,
    fill: Optional[float] = None
) -> npt.ArrayLike:
    """
    Compute the classical magnitude for a given flux measurement and photometric zeropoint.

    Parameters:
    -----------
    flux : array-like
        Input fluxes of objects.
    zeropoint : array-like
        Photometric zeropoint (per object).
    fill : array-like
        Fill value to use for objects with flux <= 0, defaults to NaN.
    
    Returns:
    --------
    mag : array-like
        Classical magnitude for objects.
    """
    if fill is None:
        fill = np.nan
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        mag = np.where(flux > 0.0, -pogson * np.log(flux) + zeropoint, fill)
    return mag


def compute_classic_magnitude_error(
    flux: npt.ArrayLike,
    flux_error: npt.ArrayLike,
    fill: Optional[float] = None
) -> npt.ArrayLike:
    """
    Compute the error of the classical magnitude for a given flux measurement and error.

    Parameters:
    -----------
    flux : array-like
        Input fluxes of objects.
    flux_error : array-like
        Input flux errors of objects.
    fill : array-like
        Fill value to use for objects with flux <= 0, defaults to NaN.
    
    Returns:
    --------
    mag : array-like
        Classical magnitude error for objects.
    """
    if fill is None:
        fill = np.nan
    mag_err = np.where(flux > 0.0, pogson * flux_error / flux, fill)
    return mag_err


def compute_magnitude(
    norm_flux: npt.ArrayLike,
    b: npt.ArrayLike
) -> npt.ArrayLike:
    """
    Compute the hyperbolic magnitude for a given flux measurement and smoothing parameter b.

    Parameters:
    -----------
    norm_flux : array-like
        Normalised fluxes (flux divided by reference flux) of objects.
    b : array-like
        Smoothing parameter (per object).
    
    Returns:
    --------
    mag : array-like
        Hyperbolic magnitude for objects.
    """
    return -pogson * (np.arcsinh(0.5 * norm_flux / b) + np.log(b))


def compute_magnitude_error(
    norm_flux: npt.ArrayLike,
    b: npt.ArrayLike,
    norm_flux_err: npt.ArrayLike
) -> npt.ArrayLike:
    """
    Compute the error of the hyperbolic magnitude for a given flux measurement and error, and
    smoothing parameter.

    Parameters:
    -----------
    norm_flux : array-like
        Normalised fluxes (flux divided by reference flux) of objects.
    b : array-like
        Smoothing parameter (per object).
    norm_flux_err : array-like
        Error of the normalised fluxes of objects.
    
    Returns:
    --------
    error : array-like
        Hyperbolic magnitude error for objects.
    """
    num = pogson * norm_flux_err
    denom = np.sqrt(norm_flux**2 + 4 * b**2)
    return num / denom


def compute_flux_stats(
    flux: npt.NDArray,
    error: npt.NDArray,
    fields: npt.NDArray,
    magnitudes: Optional[npt.NDArray] = None,
    zeropoint: Optional[npt.ArrayLike] = None,
    is_good: Optional[npt.NDArray] = None
) -> pd.DataFrame:
    """
    Compute the median flux error, photometric zeropoint, and reference flux per field (e.g. spatial
    regions, telescope pointings, etc.). The field identifiers can be set to a single value to
    compute the statistics globally. 

    Parameters:
    -----------
    flux : array-like
        Input fluxes of objects.
    error : array-like
        Input flux errors of objects.
    fields : array-like
        List of field identifiers that defines the field membership of each object.
    magnitudes : array-like
        Magnitudes of objects, used optionally to compute photometric zeropoints per object.
    zeropoint : array-like
        If no magnitudes are provided, use this photometric zeropoint (per object).
    is_good : array-like of bool
        Mask to exclude objects from computation by setting their mask value to False.

    Returns:
    --------
    stats : pd.DataFrame
        Dataframe with field identifier as index. The columns list the median flux error, the
        photometric zeropoint, and reference flux per field.
    """
    if magnitudes is None and zeropoint is None:
        raise ValueError("either magnitudes or a zeropoint must be provided")
    if is_good is None:
        is_good = np.ones(len(error), dtype="bool")
    df = pd.DataFrame({
        Keys.field: fields,
        Keys.flux_err: np.where(is_good, error, np.nan)})
    if zeropoint is None:
        df[Keys.zp] = estimate_zp(
            np.where(is_good, magnitudes, np.nan),
            np.where(is_good, flux, np.nan))
    else:
        df[Keys.zp] = zeropoint
    stats = df.groupby(Keys.field).agg(np.nanmedian)
    stats.index.name = Keys.field
    stats[Keys.ref_flux] = ref_flux_from_zp(stats[Keys.zp])
    return stats


def fill_missing_stats(stats: pd.DataFrame) -> pd.DataFrame:
    """
    Substitute missing values of per-field statistics (see `compute_flux_stats`) with the mean over
    all fields. Returns a new copy of the input statistics.
    """
    fixed_stats = stats.copy()
    mean = stats.mean()
    row_good = np.isfinite(stats).all(axis=1).to_numpy()
    if not np.all(row_good):
        fields = ", ".join(str(r) for r in stats.index[~row_good])
        logger.warn(
            f"replaced statistics of fields with global mean: {fields}")
        for col in stats.columns:
            fixed_stats[col] = np.where(row_good, stats[col], mean[col])
    return fixed_stats


def compute_flux_error_target(adapt_stats, target_zp=0.0):
    adapt_errors = convert_flux(
        adapt_stats[Keys.flux_err],
        adapt_stats[Keys.zp], target_zp)
    adapt_errors = adapt_errors.groupby(
        Keys.filter).agg(np.nanmedian)
    return adapt_errors


def adapt_flux(flux, error, stats, adapt_error, fields, is_good=None):
    if is_good is None:
        is_good = np.ones(len(error), dtype="bool")
    # convert the targeted flux error to the correct zeropoint (originally ZP=0.0)
    adapt_error = adapt_error * np.exp(stats[Keys.zp] / pogson)
    # compute the additional noise that must be added to the flux
    add_variance = adapt_error**2 - stats[Keys.flux_err]**2
    is_modified = add_variance > 0.0  # find fields that cannot be adapted
    add_variance[:] = np.where(is_modified, add_variance, 0.0)
    if not is_modified.all():
        n_unchanged = np.count_nonzero(~is_modified)
        logger.warn(f"flux in {n_unchanged} fields too large to be adapted")
    # compute standard deviation for each object
    add_sigma = fields_to_source(
        np.sqrt(add_variance), fields, index=flux.index)
    # updated the flux and flux error
    new_flux = np.where(
        is_good, np.random.normal(flux, add_sigma), flux)
    new_error = np.where(
        is_good, np.sqrt(error**2 + add_sigma**2), error)
    return new_flux, new_error



class Keys:
    filter = "filter"
    field = "field ID"
    b = "b relative"
    b_abs = "b absolute"
    ref_flux = "ref. flux"
    zp = "zeropoint"
    flux_err = "flux error"

from .magnitudes import (
    pogson, fields_to_source, estimate_b,
    ref_flux_from_zp, zp_from_ref_flux, estimate_zp, convert_flux,
    compute_classic_magnitude, compute_classic_magnitude_error,
    compute_magnitude, compute_magnitude_error,
    compute_flux_stats, fill_missing_stats,
    compute_flux_error_target, adapt_flux)



import argparse
import json
import os

import astropandas as apd
import numpy as np
import pandas as pd

from . import Keys
from .logger import logger


error_suffix = "_err"


class DumpAction(argparse.Action):

    def __call__(self, parser, namespace, values, option_string=None):
        default = json.dumps({
            "filter_name (e.g. 'r')": {
                "outname": "# output column name (suffix appended for "
                          f"errors: {error_suffix})",
                "flux": "# name of colum with flux",
                "error": "# name of column with flux error",
                "magnitude": "# name of column with magnitudes (value ignored "
                             "if --zeropoint is given)",
            }
        }, indent=4)
        print(default)
        parser.exit()


class LoadConfig:

    def __init__(self, args):
        self.verbose = args.verbose
        # get the commandline arguments
        self.infile = args.infile
        self.hdu = args.hdu
        self.outfile = args.outfile
        self.fields = args.fields
        # read the configuration file
        logger.info(f"reading configuration from {args.config}")
        with open(args.config) as f:
            data = json.load(f)
            self.filters = list(data.keys())
            self.outname = {
                fname: config["outname"] for fname, config in data.items()}
            self.flux = {
                fname: config["flux"] for fname, config in data.items()}
            self.error = {
                fname: config["error"] for fname, config in data.items()}
            self.magnitude = {
                fname: config["magnitude"] for fname, config in data.items()}
            self._config = data  # store for later use
        self.zeropoint = None  # default, overwritten by subclasses

    def get(self, filter_name, key):
        return self._config[filter_name][key]

    def set(self, filter_name, key, value):
        self._config[filter_name][key] = value

    def load_input(self):
        logger.info(f"reading data from {self.infile}")
        return apd.read_fits(self.infile, hdu=self.hdu)

    def get_fields(self, df):
        try:
            if self.fields is None:
                return np.zeros(len(df), dtype=np.float)
            else:
                return df[self.fields].to_numpy()
        except KeyError as e:
            raise KeyError(f"fields column {e} not found")

    def get_fluxes(self, df):
        try:
            return {key: df[val] for key, val in self.flux.items()}
        except KeyError as e:
            raise KeyError(f"flux column {e} not found")

    def get_errors(self, df):
        try:
            return {key: df[val] for key, val in self.error.items()}
        except KeyError as e:
            raise KeyError(f"flux error column {e} not found")

    def get_magnitudes(self, df):
        try:
            if self.zeropoint is None:
                return {key: df[val] for key, val in self.magnitude.items()}
            else:
                return None
        except KeyError as e:
            raise KeyError(f"magnitude column {e} not found")

    def verify_filters(self, df_with_filter_col):
        df_filters = set(
            df_with_filter_col.index.get_level_values(Keys.filter).unique())
        conf_filters = set(self.filters)
        if df_filters != conf_filters:
            message = "filter set does not match the configuration file"
            logger.error(message)
            raise ValueError(message)


class LoadConfigSmooting(LoadConfig):

    def __init__(self, args):
        super().__init__(args)
        # check the additional commandline arguments
        self.zeropoint = args.zeropoint
        if self.zeropoint is not None:
            self.magnitude = None
        self.adapt = args.adapt
        self.suffix = args.suffix
        self.adapt_file = args.adapt_file

    def KiDS_aware_colname(self, flux_col_name):
        if "GAAP" in flux_col_name:
            colname = flux_col_name.replace("GAAP", "GAAPadapt")
        else:
            colname = flux_col_name + self.suffix
        return colname

    def add_column_and_update(self, data, values, filt, which):
        key = self.KiDS_aware_colname(self.get(filt, which))
        logger.info(f"adding {which} column: {key}")
        data[key] = values.astype(np.float32)  # default 64 bit is overkill
        self.set(filt, which, key)

    def load_adapt(self):
        if self.adapt is not None:
            logger.info(f"reading external statistics from {self.adapt}")
            data = pd.read_csv(self.adapt, index_col=[Keys.filter, Keys.field])
            self.verify_filters(data)
            return data
        else:
            return None

    def write_output(self, data):
        if self.adapt_file is None:
            fpath = "_adapted".join(os.path.splitext(self.infile))
        else:
            fpath = self.adapt_file
        logger.info(f"writing adapted table data to {fpath}")
        apd.to_fits(data, fpath)

    def write_stats(self, data):
        logger.info(f"writing statistics to {self.outfile}")
        data.to_csv(self.outfile)


class LoadConfigMagnitudes(LoadConfig):

    def __init__(self, args):
        super().__init__(args)
        # check the additional commandline arguments
        self.stats = args.stats
        self.smoothing = args.smoothing
        self.b_global = args.b_global
        self.plot = args.plot

    def load_stats(self):
        logger.info(f"reading statistics from {self.stats}")
        data = pd.read_csv(self.stats, index_col=[Keys.filter, Keys.field])
        self.verify_filters(data)
        return data

    def load_smoothing(self):
        fpath = self.smoothing if self.smoothing is not None else self.stats
        logger.info(f"reading statistics from {fpath}")
        data = pd.read_csv(fpath, index_col=[Keys.filter, Keys.field])
        self.verify_filters(data)
        return data

    @staticmethod
    def KiDS_aware_error_colname(mag_col_name):
        if mag_col_name.startswith("HMAG_"):
            error_colname = f"HMAGERR_{mag_col_name[5:]}"
        else:
            error_colname = mag_col_name + error_suffix
        return error_colname

    def write_output(self, data):
        logger.info(f"writing table data to {self.outfile}")
        apd.to_fits(data, self.outfile)



import os

import numpy as np
import pandas as pd
import scipy.stats
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

from . import Keys
from .logger import logger
from .magnitudes import compute_classic_magnitude
import hyperbolic


class SparseSampler:

    def __init__(self, total_length, samples=50000):
        sparse = np.arange(total_length)
        np.random.shuffle(sparse)
        self.sparse = sparse[:min(total_length, samples)]

    def apply(self, data, mask=None):
        try:
            data = data.to_numpy()
        except AttributeError:
            pass
        if mask is not None:
            data = data[mask]
        return data[self.sparse]


def add_left_edge(centers, edges):
    try:
        centers = centers.to_numpy()
    except AttributeError:
        pass
    try:
        edges = edges.to_numpy()
    except AttributeError:
        pass
    return np.append(edges[0], centers)


def compute_prediction(sn, b):
    # calculate normalised flux from relation of b and S/N
    norm_flux = sn * b / np.sqrt(hyperbolic.pogson)
    return hyperbolic.compute_magnitude(norm_flux, b)


def fig_add_xlabel(axes, label, offset=0):
    ncols = axes.shape[1]
    for i in range(axes.size+offset-ncols, axes.size+offset):
        axes.flatten()[i].set_xlabel(label)


def fig_add_ylabel(axes, label):
    for ax in axes[:, 0]:
        ax.set_ylabel(label)


class PlotLims:

    def __init__(self):
        self.lower = np.inf
        self.upper = -np.inf

    def update(self, lower=None, upper=None):
        if lower is not None:
            self.lower = min(lower, self.lower)
        if upper is not None:
            self.upper = max(upper, self.upper)

    def get(self, reverse=False):
        if reverse:
            return (self.upper, self.lower)
        else:
            return (self.lower, self.upper)


class Plotter:

    _magnitudes = None

    scatter_style = {
        "edgecolor": "none", "s": 3, "marker": ".",
        "alpha": 0.1, "rasterized": True}

    def __init__(self, config, data, stats, b):
        self.config = config
        self.data = data
        self.stats = stats
        self.b = b

    def make_figure(self, size=2.5, sharex=True, sharey=True, n_plot_offset=0):
        n_plots = len(self.config.filters) + n_plot_offset
        ncols = min(3, n_plots)
        nrows = n_plots // ncols
        if nrows * ncols < n_plots:
            nrows += 1
        fig, axes = plt.subplots(
            nrows, ncols, figsize=(0.5 + size*ncols, 0.5 + size*ncols),
            sharex=sharex, sharey=sharey)
        for i, ax in enumerate(axes.flatten()):
            if i >= n_plots:
                ax.get_xaxis().set_visible(False)
                ax.get_yaxis().set_visible(False)
                plt.axis("off")
            else:
                for pos in ["top", "right"]:
                    ax.spines[pos].set_visible(False)
                ax.grid(alpha=0.33)
        return fig, axes

    def __enter__(self, *args, **kwargs):
        fname = os.path.splitext(self.config.outfile)[0] + ".pdf"
        logger.info(f"plotting to {fname}")
        self._backend = PdfPages(fname)
        return self

    def __exit__(self, *args, **kwargs):
        self._backend.close()

    def add_fig(self, fig):
        self._backend.savefig(fig)

    def plot_b(self, smoothing):
        # collect the b values
        b_median = {}
        b_all = {}
        for filt, data in smoothing.groupby(Keys.filter):
            b_all[filt] = data[Keys.b_abs].to_numpy()
            b_median[filt] = np.nanmedian(b_all[filt])
        # make figure
        fig, axes = self.make_figure(sharex=False)
        for i, filt in enumerate(self.config.filters):
            ax = axes.flatten()[i]
            _, bins, patches = ax.hist(b_all[filt], histtype="step")
            hand2 = ax.axvline(x=b_median[filt], color="k")
            # decorate
            ax.annotate(
                filt, (0.95, 0.95), xycoords="axes fraction",
                ha="right", va="top")
        fig_add_xlabel(axes, "Smoothing $b^\prime$")
        # add legend and fix layout
        fig.legend(
            handles=[patches[0], hand2],
            labels=["fields", "median (fiducial)"],
            ncol=2, frameon=False)
        fig.tight_layout()
        fig.subplots_adjust(top=0.94)
        self.add_fig(fig)

    def get_magnitudes(self):
        if self._magnitudes is not None:
            return self._magnitudes
        magnitudes = self.config.get_magnitudes(self.data)
        # compute classic magnitudes as fallback from fluxes and zeropoints
        if magnitudes is None:
            fields = self.config.get_fields(self.data)
            fluxes = self.config.get_fluxes(self.data)
            magnitudes = {}
            for filt in self.config.filters:
                zeropoint = self.stats[Keys.zp].loc[fields]
                zeropoint.index = fluxes[filt].index
                magnitudes = compute_classic_magnitude(fluxes[filt], zeropoint)
        self._magnitudes = magnitudes
        return magnitudes

    def plot_magnitudes(self):
        # get the required data
        fluxes = self.config.get_fluxes(self.data)
        errors = self.config.get_errors(self.data)
        magnitudes = self.get_magnitudes()
        # make figure
        xlims = [-3, 33]
        ylims = PlotLims()
        fig, axes = self.make_figure()
        for i, filt in enumerate(self.config.filters):
            # select all observed objects and create a sparse sampling
            ax = axes.flatten()[i]
            is_good = errors[filt] > 0.0
            sparse = SparseSampler(np.count_nonzero(is_good))
            SN = (sparse.apply(fluxes[filt], mask=is_good) /
                  sparse.apply(errors[filt], mask=is_good))
            # add prediction
            SN_theory = np.linspace(*xlims, 50)
            mag_theory = compute_prediction(SN_theory, self.b[filt])
            hand_theory = ax.plot(
                SN_theory, mag_theory, color="k", lw=0.7, ls="--", zorder=2)[0]
            ylims.update(  # update limits from y=mag(-b) to y=max(min)
                lower=mag_theory.min(),
                upper=hyperbolic.compute_magnitude(
                    -self.b[filt], self.b[filt]))
            # add classical magnitudes
            mags = sparse.apply(magnitudes[filt], mask=is_good)
            hand1 = ax.scatter(
                SN, mags, c="C0", zorder=-1, **self.scatter_style)
            # add hyperbolic magnitudes
            key_mag = self.config.outname[filt]
            mags = sparse.apply(self.data[key_mag], mask=is_good)
            hand2 = ax.scatter(
                SN, mags, c="C3", **self.scatter_style)
            # decorate
            ax.axhline(
                y=hyperbolic.compute_magnitude(0.0, self.b[filt]),
                color="k", lw=0.4)
            ax.axvline(x=0.0, color="k", lw=0.25)
            ax.annotate(
                filt, (0.1, 0.95), xycoords="axes fraction",
                ha="left", va="top")
            ax.set_xlim(*xlims)
            ax.set_ylim(*ylims.get(reverse=True))
        fig_add_xlabel(axes, "Signal-to-noise")
        fig_add_ylabel(axes, "Magnitude")
        # add legend and fix layout
        fig.legend(
            handles=[hand_theory, hand1, hand2],
            labels=["prediction", "classical", "hyperbolic"],
            markerscale=8, ncol=3, frameon=False)
        fig.tight_layout()
        fig.subplots_adjust(top=0.94)
        self.add_fig(fig)

    def plot_magnitude_distribution(self):
        # get the required data
        errors = self.config.get_errors(self.data)
        magnitudes = self.get_magnitudes()
        # make figure
        bins = np.arange(5, 40, 0.2)
        xlims = PlotLims()
        fig, axes = self.make_figure()
        for i, filt in enumerate(self.config.filters):
            # select all observed objects and create a sparse sampling
            ax = axes.flatten()[i]
            is_good = errors[filt] > 0.0
            has_mag = magnitudes[filt] < 90.0
            # add classical magnitudes
            c, _, patches1 = ax.hist(
                magnitudes[filt][is_good & has_mag], bins,
                log=True, color="C0", histtype="step")
            # add hyperbolic magnitudes
            key = self.config.outname[filt]
            c, _, patches2 = ax.hist(
                self.data[key][is_good], bins,
                log=True, color="C3", histtype="step")
            # decorate
            zeroflux = hyperbolic.compute_magnitude(0.0, self.b[filt])
            hand0 = ax.axvline(x=zeroflux, color="k", lw=0.7, ls="--")
            xlims.update(self.data[key][is_good].min(), zeroflux + 5)
            ax.annotate(
                filt, (0.05, 0.95), xycoords="axes fraction",
                ha="left", va="top")
            ax.set_xlim(*xlims.get())
        fig_add_xlabel(axes, "Magnitude")
        fig_add_ylabel(axes, "Frequency")
        # add legend and fix layout
        fig.legend(
            handles=[hand0, patches1[0], patches2[0]],
            labels=["zero-flux magnitude", "classical", "hyperbolic"],
            ncol=3, frameon=False)
        fig.tight_layout()
        fig.subplots_adjust(top=0.94)
        self.add_fig(fig)


    def plot_colour_distribution(self):
        # get the required data
        errors = self.config.get_errors(self.data)
        magnitudes = self.get_magnitudes()
        # make figure
        bins = np.linspace(-1.5, 2.5, 50)
        fig, axes = self.make_figure(n_plot_offset=-1)
        for i, (filt1, filt2) in enumerate(
                zip(self.config.filters[:-1], self.config.filters[1:])):
            ax = axes.flatten()[i]
            # select all observed objects and create a sparse sampling
            mag1 = magnitudes[filt1]
            mag2 = magnitudes[filt2]
            is_good = (errors[filt1] > 0.0) & (errors[filt2] > 0.0)
            has_color = np.isfinite(mag1) & (mag1 < 90.0)
            has_color &= np.isfinite(mag2) & (mag2 < 90.0)
            mask = is_good & has_color
            # add classical colours
            mag1 = magnitudes[filt1][mask].to_numpy()
            mag2 = magnitudes[filt2][mask].to_numpy()
            idx_sort = np.argsort(mag1)
            colours = mag1[idx_sort] - mag2[idx_sort]
            hand1a = ax.hist(
                colours[:len(colours)//4], bins,
                color="C0", alpha=0.3, zorder=-1)[-1][0]
            hand1b = ax.hist(
                colours[-len(colours)//4:], bins,
                color="C0", histtype="step")[-1][0]
            # add hyperbolic colours
            key_mag1 = self.config.outname[filt1]
            key_mag2 = self.config.outname[filt2]
            mag1 = self.data[key_mag1][mask].to_numpy()
            mag2 = self.data[key_mag2][mask].to_numpy()
            idx_sort = np.argsort(mag1)
            colours = (mag1[idx_sort] - mag2[idx_sort])
            hand2a = ax.hist(
                colours[:len(colours)//4], bins,
                color="C3", alpha=0.3, zorder=-1)[-1][0]
            hand2b = ax.hist(
                colours[-len(colours)//4:], bins,
                color="C3", histtype="step")[-1][0]
            # decorate
            ax.axvline(x=0.0, color="k", lw=0.25)
            ax.annotate(
                f"{filt1}$-${filt2}", (0.1, 0.95), xycoords="axes fraction",
                ha="left", va="top")
            ax.set_xlim(bins[0], bins[-1])
        fig_add_xlabel(axes, "Colour", offset=-1)
        fig_add_ylabel(axes, "Frequency")
        # add legend and fix layout
        axes.flatten()[-1].legend(
            loc="center",
            handles=[hand1a, hand1b, hand2a, hand2b],
            labels=[
                "classical\n(bright quartile)",
                "classical\n(faint quartile)",
                "hyperbolic\n(bright quartile)",
                "hyperbolic\n(faint quartile)"],
            ncol=1, frameon=False)
        fig.tight_layout()
        fig.subplots_adjust()
        self.add_fig(fig)

    def plot_magdiff(self):
        # get the required data
        errors = self.config.get_errors(self.data)
        magnitudes = self.get_magnitudes()
        # make figure
        fig, axes = self.make_figure(sharey=False)
        xlims = PlotLims()
        for i, filt in enumerate(self.config.filters):
            # select all observed objects and create a sparse sampling
            ax = axes.flatten()[i]
            is_good = (errors[filt] > 0.0) & (magnitudes[filt] < 90.0)
            # collect the data
            key_mag = self.config.outname[filt]
            df = pd.DataFrame({
                "mag": magnitudes[filt][is_good],
                "hyp": self.data[key_mag][is_good]})
            df["diff"] = df["mag"] - df["hyp"]
            sparse = SparseSampler(np.count_nonzero(is_good))
            # add the magnitude difference
            ax.scatter(
                sparse.apply(df["mag"]), sparse.apply(df["diff"]),
                **self.scatter_style)
            # plot statistics
            bins = np.linspace(*np.percentile(df["mag"], q=[0.5, 99.5]), 20)
            centers = (bins[1:] + bins[:-1]) / 2.0
            stats = df.groupby(pd.cut(df["mag"], bins)).agg([
                np.median, scipy.stats.median_abs_deviation])
            y = stats["diff"]["median"]
            ylow = y - stats["diff"]["median_abs_deviation"]
            yhigh = y + stats["diff"]["median_abs_deviation"]
            ax.plot(
                add_left_edge(centers, bins), add_left_edge(y, y),
                color="k", lw=0.7)
            ax.plot(
                add_left_edge(centers, bins), add_left_edge(ylow, ylow),
                color="k", ls="--", lw=0.7)
            ax.plot(
                add_left_edge(centers, bins), add_left_edge(yhigh, yhigh),
                color="k", ls="--", lw=0.7)
            # decorate
            zeroflux = hyperbolic.compute_magnitude(0.0, self.b[filt])
            xlims.update(bins[0], zeroflux)
            ax.set_xlim(*xlims.get())
            ax.set_ylim(-0.2, 0.6)
            ax.annotate(
                filt, (0.1, 0.95), xycoords="axes fraction",
                ha="left", va="top")
        fig_add_xlabel(axes, "Classical")
        fig_add_ylabel(axes, "Classical $-$ hyperbolic")
        # add legend and fix layout
        fig.tight_layout()
        fig.subplots_adjust()
        self.add_fig(fig)



import setuptools
import os


with open("README.md", "r") as f:
    long_description = f.read()
with open("requirements.txt") as f:
    install_requires = [pkg.strip() for pkg in f.readlines() if pkg.strip()]
scripts = [
    os.path.join("bin", fname) for fname in os.listdir("bin")]

setuptools.setup(
    name="pz-hyperbolic-temp",
    version="0.0",
    author="Jan Luca van den Busch",
    description="Implementation of hyperbolic magnitudes (Lupton et al. 1999).",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jlvdb/hyperbolic",
    packages=setuptools.find_packages(),
    scripts=scripts,
    install_requires=install_requires)



#!/usr/bin/env python3
import argparse
from types import new_class

import numpy as np
import pandas as pd

import hyperbolic
import hyperbolic.config


parser = argparse.ArgumentParser(
    description="Compute the statistics of a photometric data catalogue that "
                "are needed to compute the smoothing factor for the "
                "hyperbolic magnitudes.",
    add_help=False)
parser.add_argument(
    "infile", metavar="infile",
    help="input FITS file")
parser.add_argument(
    "--hdu", type=int, default=1,
    help="FITS HDU index to read (default: %(default)s)")
parser.add_argument(
    "outfile", metavar="statfile",
    help="output path for data statistics")

parser.add_argument(
    "-c", "--config", required=True,
    help="JSON configuration file that specifies in- and output data "
         "(see --dump)")
parser.add_argument(
    "-f", "--fields",
    help="column name that can uniquely indentify pointings")
parser.add_argument(
    "--zeropoint", type=float,
    help="use this fixed magnitude zeropoint")

adpat = parser.add_argument_group(
    "flux adaptation",
    description="Adjust the flux noise level in the input data to match that "
                "of an external data catalogue before measuring statistics.")
adpat.add_argument(
    "-a", "--adapt",
    help="statistics file genereated with this script for the external data "
         "catalogue")
adpat.add_argument(
    "--suffix", default="_adapt",
    help="suffix that is added to the names of data columns with the adapted "
         "fluxes and flux errors (default: %(default)s)")
adpat.add_argument(
    "--adapt-file",
    help="file path where the output FITS file with added adapted fluxes is "
         "written (default derived from input: [infile]_adapted.*)")

group = parser.add_argument_group("help")
group.add_argument(
    "-h", "--help", action="help",
    help="show this help message and exit")
group.add_argument(
    "-d", "--dump", nargs=0, action=hyperbolic.config.DumpAction,
    help="dump an empty configuration file, 'filter_name' can be repeated")
parser.add_argument(
    "-v", "--verbose", action="store_true",
    help="show statistic summary per filter")


if __name__ == "__main__":
    config = hyperbolic.config.LoadConfigSmooting(parser.parse_args())
    # load the input data
    data = config.load_input()
    adapt_stats = config.load_adapt()  # statistics of the external data
    # get data columns
    fields = config.get_fields(data)
    fluxes = config.get_fluxes(data)
    errors = config.get_errors(data)
    magnitudes = config.get_magnitudes(data)

    # compute median flux error in the external catalogue for fixed ZP=0
    if config.adapt is not None:
        adapt_errors = hyperbolic.compute_flux_error_target(adapt_stats, 0.0)
        if config.verbose:  # print in nicely looking dataframe
            _df = pd.DataFrame({hyperbolic.Keys.flux_err: adapt_errors})
            print(_df.loc[config.filters])  # maintain usual order
        adapt_errors = adapt_errors.to_dict()
    else:
        adapt_errors = {filt: None for filt in config.filters}

    # compute the data statistics required for the smoothing
    all_stats = []
    for filt in config.filters:
        hyperbolic.logger.logger.info(f"processing filter {filt}")
        # mask to valid observations
        is_good = errors[filt] > 0.0

        # compute the additional noise required to match the external data
        if config.adapt is not None:
            # compute zeropoint and median flux error
            stats = hyperbolic.compute_flux_stats(
                fluxes[filt], errors[filt], fields, magnitudes[filt],
                config.zeropoint, is_good)
            stats = hyperbolic.fill_missing_stats(stats)
            zeropoint = hyperbolic.fields_to_source(
                stats[hyperbolic.Keys.zp], fields, index=fluxes[filt].index)
            # updated the flux and flux error and add to input data table
            fluxes[filt], errors[filt] = hyperbolic.adapt_flux(
                fluxes[filt], errors[filt], stats, adapt_errors[filt],
                fields, is_good)
            config.add_column_and_update(data, fluxes[filt], filt, "flux")
            config.add_column_and_update(data, errors[filt], filt, "error")
            # compute the classical adapted magnitudes
            if magnitudes[filt] is None:
                fill_mag, fill_err = None, None
            else:
                if np.isnan(magnitudes[filt]).any():
                    fill_mag, fill_err = None, None
                else:
                    fill_mag = magnitudes[filt].max()
                    fill_err = hyperbolic.compute_classic_magnitude(
                        errors[filt], zeropoint)
            mag_err = hyperbolic.compute_classic_magnitude_error(
                fluxes[filt], errors[filt], fill=fill_err)
            # update the internal values
            magnitudes[filt] = hyperbolic.compute_classic_magnitude(
                fluxes[filt], zeropoint, fill=fill_mag)
            # include them in the input data
            data[f"mag_classical_adapt_{filt}"] = magnitudes[filt]
            data[f"magerr_classical_adapt_{filt}"] = mag_err

        # compute zeropoint and median flux error
        stats = hyperbolic.compute_flux_stats(
            fluxes[filt], errors[filt], fields,
            None if magnitudes is None else magnitudes[filt],
            config.zeropoint, is_good)
        # compute b
        stats[hyperbolic.Keys.b] = hyperbolic.estimate_b(
            stats[hyperbolic.Keys.zp], stats[hyperbolic.Keys.flux_err])
        stats[hyperbolic.Keys.b_abs] = \
            stats[hyperbolic.Keys.ref_flux] * stats[hyperbolic.Keys.b]

        # collect statistics
        stats[hyperbolic.Keys.filter] = filt
        stats = stats.reset_index().set_index([
            hyperbolic.Keys.filter, hyperbolic.Keys.field])
        if config.verbose:
            print(stats)
        all_stats.append(stats)

    if config.adapt is not None:
        config.write_output(data)
    config.write_stats(pd.concat(all_stats))
